from datetime import timedelta
from typing import Annotated, Any, ClassVar

from fastapi.security import HTTPAuthorizationCredentials
from pydantic import AnyUrl, BaseModel, ConfigDict, Field


class JWTDecodeConfig(BaseModel):
    """Options passed to pyjwt's jwt.decode()"""

    audience: list[str] | None = None
    issuer: str | None = None
    leeway: float | timedelta = 0
    options: dict[str, Any] | None = None
    verify: bool | None = None


class JWKSAuthConfig(BaseModel):
    payload_field: str = "payload"
    token_field: str = "raw_token"


class JWKSConfig(BaseModel):
    url: str
    ca_cert_path: str | None = None


class JWTTokenInjectorConfig(BaseModel):
    payload_field: str = "payload"
    token_field: str = "raw_token"


class JWKSAuthCredentials[DataT: BaseModel](HTTPAuthorizationCredentials):
    payload: DataT


class JWTHeader(BaseModel):
    """JSON Web Token

    RFC: https://datatracker.ietf.org/doc/html/rfc7519
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow")

    alg: Annotated[str, Field(description="Algorithm used for signing")]
    typ: Annotated[
        str | None,
        Field(description="Type of token", examples=["JWT"]),
    ] = None
    cty: Annotated[str | None, Field(description="Content type")] = None
    kid: Annotated[str | None, Field(description="Key ID")] = None
    x5u: Annotated[str | None, Field(description="X.509 URL")] = None
    x5c: Annotated[list[str] | None, Field(description="X.509 Certificate Chain")] = (
        None
    )
    x5t: Annotated[
        str | None, Field(description="X.509 Certificate SHA-1 Thumbprint")
    ] = None
    x5t_s256: Annotated[
        str | None,
        Field(
            description="X.509 Certificate SHA-256 Thumbprint",
            validation_alias="x5t#S256",
        ),
    ] = None


class JWK(BaseModel):
    """JSON Web Key

    RFC: https://datatracker.ietf.org/doc/html/rfc7517#section-4
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow", frozen=True)

    kty: Annotated[str, Field(description="Key Type")]
    use: Annotated[str | None, Field(description="Public key use")] = None
    key_ops: Annotated[list[str] | None, Field(description="Key operations")] = None
    alg: Annotated[str | None, Field(description="Algorithm intended for the key")] = (
        None
    )
    kid: Annotated[str | None, Field(description="Key ID")] = None
    x5u: Annotated[AnyUrl | None, Field(description="X.509 URL")] = None
    x5c: Annotated[list[str] | None, Field(description="X.509 Certificate Chain")] = (
        None
    )
    x5t: Annotated[
        str | None, Field(description="X.509 Certificate SHA-1 Thumbprint")
    ] = None
    x5t_s256: Annotated[
        str | None,
        Field(
            description="X.509 Certificate SHA-256 Thumbprint",
            validation_alias="x5t#S256",
        ),
    ] = None
    n: Annotated[str | None, Field(description="Modulus for RSA keys")] = None
    e: Annotated[str | None, Field(description="Exponent for RSA keys")] = None
    crv: Annotated[str | None, Field(description="Curve for EC keys")] = None
    x: Annotated[str | None, Field(description="X coordinate for EC keys")] = None
    y: Annotated[str | None, Field(description="Y coordinate for EC keys")] = None
    d: Annotated[
        str | None, Field(description="Private exponent for RSA or EC private key")
    ] = None
    k: Annotated[str | None, Field(description="Symmetric key value")] = None


class JWKS(BaseModel):
    keys: Annotated[list[JWK], Field(min_length=1)]

    @property
    def algorithms(self) -> list[str]:
        return [key.alg for key in self.keys if key.alg is not None]



import logging
from functools import cached_property
from typing import Any, final

import httpx
import jwt
import pydantic
from cachetools import TTLCache, cached
from fastapi import HTTPException, status
from jwt import algorithms
from pydantic import BaseModel

from fastapi_jwks.models.types import JWKS, JWKSConfig, JWTDecodeConfig, JWTHeader

logger = logging.getLogger("fastapi-jwks")


@final
class JWKSValidator[DataT: BaseModel]:
    def __init__(self, decode_config: JWTDecodeConfig, jwks_config: JWKSConfig):
        self.decode_config = decode_config
        self.jwks_config = jwks_config
        self.client = self._create_client()

    def _create_client(self) -> httpx.Client:
        client_kwargs: dict[str, Any] = {}
        if self.jwks_config.ca_cert_path:
            client_kwargs["verify"] = self.jwks_config.ca_cert_path
        return httpx.Client(**client_kwargs)

    @cached(cache=TTLCache(ttl=600, maxsize=1))
    def jwks_data(self) -> JWKS:
        try:
            logger.debug("Fetching JWKS from %s", self.jwks_config.url)
            jwks_response = self.client.get(self.jwks_config.url)
            jwks_response.raise_for_status()
        except httpx.RequestError as e:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
                detail="Invalid JWKS URI",
            ) from e
        try:
            return JWKS.model_validate(jwks_response.json())
        except pydantic.ValidationError as e:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail="Invalid JWKS"
            ) from e

    @cached_property
    def __is_generic_passed(self) -> bool:
        return getattr(self, "__orig_class__", None) is not None

    def validate_token(self, token: str) -> DataT:
        if not self.__is_generic_passed:
            raise ValueError(
                "Validator needs a model as generic value to decode payload"
            )

        public_key: bytes | None = None
        try:
            header = JWTHeader.model_validate(jwt.get_unverified_header(token))
            jwks_data = self.jwks_data()
            provided_algorithms = jwks_data.algorithms
            if provided_algorithms and header.alg not in provided_algorithms:
                logger.debug(
                    f"Could not find '{header.alg}' in provided algorithms: {provided_algorithms}"
                )
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid token"
                )
            for key in jwks_data.keys:
                if key.kid == header.kid:
                    public_key = algorithms.get_default_algorithms()[
                        header.alg
                    ].from_jwk(key.model_dump(exclude_none=True))
                    break
            if public_key is None:
                logger.debug(
                    f"No public key for provided algorithm '{header.alg}' found in JWKS data"
                )
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid token"
                )
            return self.__orig_class__.__args__[0].model_validate(  # pyright: ignore[reportAttributeAccessIssue]
                # This line gets the generic value in runtime to transform it to the correct pydantic model
                jwt.decode(
                    token,
                    key=public_key,
                    **self.decode_config.model_dump(),
                    algorithms=[header.alg],
                )
            )
        except jwt.ExpiredSignatureError:
            logger.debug("Expired token", exc_info=True)
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED, detail="Token has expired"
            ) from None
        except (pydantic.ValidationError, jwt.InvalidTokenError):
            logger.debug("Invalid token", exc_info=True)
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid token"
            ) from None



from typing import final, override

from fastapi import HTTPException, Request, status
from fastapi.security.http import HTTPBase
from pydantic import BaseModel

from fastapi_jwks.models.types import JWKSAuthConfig, JWKSAuthCredentials
from fastapi_jwks.validators import JWKSValidator

UNAUTHORIZED_ERROR = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail="Invalid authorization token",
)


@final
class JWKSAuth[DataT: BaseModel](HTTPBase):
    def __init__(
        self,
        jwks_validator: JWKSValidator[DataT],
        config: JWKSAuthConfig | None = None,
        auth_header: str = "Authorization",
        auth_scheme: str = "Bearer",
        scheme_name: str | None = None,
    ):
        self.config = config or JWKSAuthConfig()
        self.jwks_validator = jwks_validator
        self.auth_header = auth_header
        self.auth_scheme = auth_scheme.lower()
        super().__init__(
            scheme=self.auth_scheme, scheme_name=scheme_name, auto_error=False
        )

    @override
    async def __call__(self, request: Request) -> JWKSAuthCredentials[DataT]:
        authorization = request.headers.get(self.auth_header)
        if not authorization:
            raise UNAUTHORIZED_ERROR

        try:
            scheme, _, token = authorization.partition(" ")
            if scheme.lower() != self.auth_scheme:
                raise UNAUTHORIZED_ERROR
        except ValueError as e:
            raise UNAUTHORIZED_ERROR from e

        try:
            payload = self.jwks_validator.validate_token(token)
            setattr(request.state, self.config.payload_field, payload)
            setattr(request.state, self.config.token_field, token)
        except Exception as e:
            raise UNAUTHORIZED_ERROR from e

        return JWKSAuthCredentials[DataT](
            scheme=scheme, credentials=token, payload=payload
        )



from pydantic import BaseModel
from starlette.requests import Request

from fastapi_jwks.models.types import JWTTokenInjectorConfig


class JWTTokenInjector[DataT: BaseModel]:
    def __init__(self, config: JWTTokenInjectorConfig | None = None):
        self.config: JWTTokenInjectorConfig = config or JWTTokenInjectorConfig()

    async def __call__(self, request: Request) -> DataT:
        return getattr(request.state, self.config.payload_field)


class JWTRawTokenInjector:
    def __init__(self, config: JWTTokenInjectorConfig | None = None):
        self.config: JWTTokenInjectorConfig = config or JWTTokenInjectorConfig()

    async def __call__(self, request: Request) -> str:
        return getattr(request.state, self.config.token_field)
